/*
    JScoord - National Grid Reference (OSGB) to Latitude, Longitude and UTM coordinate conversion
    Copyright © 2005 Jonathan Stott, 2009,2016 Harry Whitfield

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    This version of JScoord is a modified and enhanced version of JScoord 1.1.1 by Jonathan Stott.
    See the notice immediately below and the licensing information at
    http://www.jstott.me.uk/jscoord/

    JScoord - version 3.0
    15 November, 2016
    Copyright © 2005 Jonathan Stott, 2009,2016 Harry Whitfield
    mailto:g6auc@arrl.net
*/

//--------------------------------------------------------------------------
// JScoord
// jscoord.js
//
// (c) 2005 Jonathan Stott
//
// Created on 21-Dec-2005
//
// 1.1.1 - 16 Jan 2006
//  - Fixed radix bug in getOSRefFromSixFigureReference that would return
//    the incorrect reference if either the northing or the easting started
//    with a leading zero.
// 1.1 - 23 Dec 2005
//  - Added getOSRefFromSixFigureReference function.
// 1.0 - 11 Aug 2005
//  - Initial version created from PHPcoord v1.1
//--------------------------------------------------------------------------

/*
    JScoord is a global object which provides a number of functions for coordinate conversion.

    JScoord.OSRef(easting, northing) is a constructor of OSGB grid reference objects.

        var osRef1 = new JScoord.OSRef(651409.903, 313177.270);

    JScoord.OSRef.toString() converts an OSRef to a string.

        var s = osRef1.toString(); // "(651409.903, 313177.27)"

    JScoord.OSRef.toOSString(figures) converts an OSRef to an alphanumeric reference

        var s = osRef1.toOSString(10); // "TG5140913177"

    JScoord.OSRef.toSixFigureString() // legacy code - use toOSString(6) instead

        var s = osRef1.toSixFigureString(); // "TG514131"

    JScoord.OSRef.toLatLng() converts an OSRef object to a position object

        var latLng1 = osRef1.toLatLng(); // (52.65757030181147, 1.717921581003804)

    JScoord.getOSRefFromOSString(osString) converts an OS string
                                                   (two letters and from 2 to 10 figures)
                                                   to an OSGB grid references object.

        var osRef2 = JScoord.getOSRefFromOSString("TG5140913177");

    JScoord.getOSRefFromSixFigureReference(osString) converts an OS string
                                                      (two letters and 6 figures)
                                                      to an OSGB grid references object.

        var osRef2 = JScoord.getOSRefFromSixFigureReference("TG514131"); // legacy code

    JScoord.getLatLngFromLocator(locator) converts an IARU locator to a position object.

        var latLng3 = JScoord.getLatLngFromLocator("IO91WL49");

    ---

    JScoord.LatLng(lat, lng) is a constructor of position objects.

        var newYork = new JScoord.LatLng(40.718119, -73.995667); // New York
        var london = new JScoord.LatLng(51.499981, -0.125313); // London

    JScoord.LatLng.toString() converts a position to a string.

        var s = newYork.toString(); // "(40.718119, -73.995667)"

    JScoord.LatLng.toLocator() converts a position to an IARU locator.

        var s = london.toLocator(); // "IO91WL49"

    JScoord.LatLng.distance(to) computes the distance between two positions.

        var dist = newYork.distance(london); // 5596.5392661564265   (km)

    JScoord.LatLng.osgb36ToWGS84() converts an OSGB36 position to a WGS84 position.

        var london1 = london.osgb36ToWGS84();

    JScoord.LatLng.wgs84ToOSGB36() converts a WGS84 position to an OSGB36 position.

        var london2 = london1.wgs84ToOSGB36();

    JScoord.LatLng.toOSRef() converts a position to an OSGB grid reference object

        var osRef3 = london.toOSRef();

    JScoord.LatLng.toUTMRef() converts a position to a UTM object

        var utmRef1 = newYork.toUTMRef();

    ---

    JScoord.UTMRef(easting, northing, latZone, lngZone) is a constructor of UTM objects.

        var utmRef2 = new JScoord.UTMRef(621160.98, 3349893.53, "R", 14);

    JScoord.UTMRef.toString() converts a UTM object to a string.

        var s = utmRef2.toString(); // "14R 621160.98 3349893.53"

    JScoord.UTMRef.toLatLng() converts a UTM object to a position.

        var latLng2 = utmRef2.toLatLng(); // (30.274673137173487, -97.74033148749622)

*/

/*jslint bitwise, for, multivar, this */

/*property
    LatLng, OSRef, PI, UTMRef, atan, charAt, charCodeAt, cos, distance,
    easting, ecc, floor, fromCharCode, getLatLngFromLocator,
    getOSRefFromOSString, getOSRefFromSixFigureReference, indexOf, isNaN, join,
    lat, latZone, length, lng, lngZone, maj, match, min, northing,
    osgb36ToWGS84, pow, prototype, round, sin, sqrt, substring, toDMSString,
    toFixed, toLatLng, toLocator, toOSRef, toOSString, toSixFigureString,
    toString, toUTMRef, toUpperCase, wgs84ToOSGB36
*/

"use strict";

var JScoord = (function () {

 // ================================================== Mathematical Functions

    function sinSquared(x) {
        var r = Math.sin(x);
        return r * r;
    }

    function sec(x) {
        return 1.0 / Math.cos(x);
    }

    function deg2rad(x) {
        return x * (Math.PI / 180);
    }

    function rad2deg(x) {
        return x * (180 / Math.PI);
    }

 // ========================================================= Other Functions

 /**
     *  Work out the UTM latitude zone from the latitude
     *
     * @param latitude
     * @return
     */
    var getUTMLatitudeZoneLetter = function (latitude) {
        var letters = "CDEFGHJKLMNPQRSTUVWX";
        if ((80 > latitude) && (latitude >= -80)) {
            return letters[Math.floor((latitude + 80) / 8)];
        }
        if ((84 >= latitude) && (latitude >= 80)) {
            return "X";
        }
        return "Z";
    };

 // =================================================================== OSRef

    var OSRef = function (easting, northing) {
        this.easting = easting;
        this.northing = northing;
    };

    OSRef.prototype.toString = function OSRefToString() {
        return "(" + this.easting + ", " + this.northing + ")";
    };

    function osRefOfLetterPair(pair) {
        pair = pair.toUpperCase();

        var error = new OSRef(NaN, NaN);

        var first = "HNSOT";
        var E1 = [0, 0, 0, 500000, 500000];
        var N1 = [1000000, 500000, 0, 500000, 0];

        var idx = first.indexOf(pair[0]);
        if (idx < 0) {
            return error;
        }
        var easting = E1[idx];
        var northing = N1[idx];

        var second = "VWXYZQRSTULMNOPFGHJKABCDE";
        var valid = [
        	"PTUYZ",
            "ABCDFGHJKLMNORSTUWXYZ",
            "CDEGHJKMNOPRSTUVWXYZ",
            "V",
            "AFGLMQRV"
        ];

        var idx2 = valid[idx].indexOf(pair[1]);
        if (idx2 < 0) {
            return error;
        }

        idx = second.indexOf(pair[1]);
        if (idx < 0) {
            return error; // should not happen!
        }

        var x = idx % 5;
        easting += 100000 * x;

        var y = Math.round((idx - x) / 5);
        northing += 100000 * y;
        return new OSRef(easting, northing);
    }

    function letterPairOfOSRef(osRef) {
        var easting = Math.floor(osRef.easting);
        var northing = Math.floor(osRef.northing);

        var idx = Math.floor(easting / 500000);
        var idy = Math.floor(northing / 500000);

        if ((idx < 0) || (idx > 1)) {
            return "";
        }
        if ((idy < 0) || (idx > 2)) {
            return "";
        }

        var first = [
        	["S", "N", "H"], ["T", "O", "J"]
        ];
        var pair = first[idx][idy];

        easting -= 500000 * idx;
        northing -= 500000 * idy;

        idx = Math.floor(easting / 100000);
        idy = Math.floor(northing / 100000);

        var second = "VWXYZQRSTULMNOPFGHJKABCDE";

        pair += second[5 * idy + idx];
        var G = osRefOfLetterPair(pair);

        if (Number.isNaN(G.easting)) {
            return "";
        }
        return pair;
    }

    function OSRefToOSString(figures) { // figures 2, 4, 6, 8, 10
        var easting = this.easting;
        var northing = this.northing;
        var tmp;

        tmp = Math.floor(easting) % 100000;
        easting = tmp.toString();
        tmp = Math.floor(northing) % 100000;
        northing = tmp.toString();

        while (easting.length < 5) {
            easting = "0" + easting;
        }
        while (northing.length < 5) {
            northing = "0" + northing;
        }

        figures = figures || 6; // default to 6 figures if undefined

        if ((figures < 2) || (figures > 10)) {
            figures = 6;
        }

        if ((figures % 2) !== 0) {
            figures = 6;
        }

        figures = figures >> 1;

        easting = easting.substring(0, figures);
        northing = northing.substring(0, figures);

        var pair = letterPairOfOSRef(this);

        if (pair !== "") {
            return pair + easting + northing;
        }
        return "Invalid Reference";
    }

    OSRef.prototype.toOSString = OSRefToOSString;

    OSRef.prototype.toSixFigureString = OSRefToOSString;

 /**
     * Take a string formatted as an alphanumeric OS grid reference (e.g.
     * "TG514131") and return a reference to an OSRef object that represents
     * that grid reference. The first character must be H, N, S, O or T.
     * The second character can be any uppercase character from A through Z
     * excluding I.
     *
     * @param osString
     * @return
     */
    var getOSRefFromOSString = function (osString) {
        var lookFor = /^([HNSOT])([VWXYZQRSTULMNOPFGHJKABCDE])(\d{2,10})$/i;
        var found = osString.match(lookFor);
        if ((found === null) || ((found[3].length % 2) !== 0)) {
            return new OSRef(NaN, NaN);
        }
        var pair = osString.substring(0, 2);
        var rest = osString.substring(2);
        var len = rest.length >> 1;
        var easting = rest.substring(0, len);
        var northing = rest.substring(len);
        while (easting.length < 5) {
            easting += "0";
        }
        while (northing.length < 5) {
            northing += "0";
        }
        easting = Number(easting);
        northing = Number(northing);
        var G = osRefOfLetterPair(pair);
        easting += G.easting;
        northing += G.northing;
        return new OSRef(easting, northing);
    };

 // ================================================================== LatLng

    var LatLng = function (lat, lng) {
        this.lat = lat;
        this.lng = lng;
    };

    LatLng.prototype.toString = function LatLngToString() {
        return "(" + this.lat + ", " + this.lng + ")";
    };

    function dms(angle, dir) { // convert angle to DD MM SS.SSSS format
        var d, m, s, ns, digits;

        ns = dir[0];
        if (angle < 0) {
            angle = -angle;
            ns = dir[1];
        }
        d = Math.floor(angle);
        angle = 60 * (angle - d);
        m = Math.floor(angle);
        s = 60 * (angle - m);

        digits = (
        	dir === "EW"
        	? 3
        	: 2
        );

        d = String(d);
        while (d.length < digits) {
            d = "0" + d;
        }

        m = String(m);
        if (m.length < 2) {
            m = "0" + m;
        }

        s = s.toFixed(4);
        if (s.length < 7) {
            s = "0" + s;
        }

        return d + " " + m + " " + s + " " + ns;
    }

    LatLng.prototype.toDMSString = function LatLngToDMSString() {
        return "[" + dms(this.lat, "NS") + ", " + dms(this.lng, "EW") + "]";
    };

    LatLng.prototype.distance = function LatLngDistance(to) {
        var er = 6366.707;

        var latFrom = deg2rad(this.lat);
        var latTo = deg2rad(to.lat);
        var lngFrom = deg2rad(this.lng);
        var lngTo = deg2rad(to.lng);

        var sinLatFrom = Math.sin(latFrom);

        var x1 = er * Math.cos(lngFrom) * sinLatFrom;
        var y1 = er * Math.sin(lngFrom) * sinLatFrom;
        var z1 = er * Math.cos(latFrom);

        var sinLatTo = Math.sin(latTo);

        var x2 = er * Math.cos(lngTo) * sinLatTo;
        var y2 = er * Math.sin(lngTo) * sinLatTo;
        var z2 = er * Math.cos(latTo);

        var d = Math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2) + (z1 - z2) * (z1 - z2));

        return d;
    };

    LatLng.prototype.toLocator = function LatLngToLocator() {
        var d, e, i, loc = [], LB = "AA00AA00";

        if ((this.lat < -90) || (this.lat > 90)) {
            return "";
        }

        if ((this.lng < -180) || (this.lng > 180)) {
            return "";
        }

        d = this.lat + 90.0;
        loc[1] = Math.floor(d / 10.0);
        d = d - 10.0 * loc[1];
        loc[3] = Math.floor(d);
        d = d - loc[3];
        loc[5] = Math.floor(24.0 * d);
        d = 24.0 * d - loc[5];
        loc[7] = Math.floor(10.0 * d);

        e = this.lng / 2 + 90.0;
        loc[0] = Math.floor(e / 10.0);
        e = e - 10.0 * loc[0];
        loc[2] = Math.floor(e);
        e = e - loc[2];
        loc[4] = Math.floor(24.0 * e);
        e = 24.0 * e - loc[4];
        loc[6] = Math.floor(10.0 * e);

        var s = [];
        for (i = 0; i < 8; i += 1) {
            s[i] = String.fromCharCode(loc[i] + LB[i].charCodeAt(0));
        }

        if (s[1] === "S") { // latitude 90 rounded down
            s[1] = "R";
            s[3] = "9";
            s[5] = "X";
            s[7] = "9";
        }
        if (s[0] === "S") { // longitude 180 maps to -180
            s[0] = "A";
        }
        return s.join("");
    };

    function getLatLngFromLocator(locator) {
        var LB = "AA00AA00", UB = "RR99XX99", loc = [], i, c;

        if (locator.length === 6) {
            locator += "55";
        }
        if (locator.length !== 8) {
            return new LatLng(NaN, NaN);
        }

        locator = locator.toUpperCase();

        for (i = 0; i < 8; i += 1) {
            c = locator[i];
            if ((LB[i] <= c) && (c <= UB[i])) {
                loc[i] = c.charCodeAt(0) - LB[i].charCodeAt(0);
            } else {
                return new LatLng(NaN, NaN);
            }
        }

        var latitude = 10.0 * loc[1] + loc[3] - 90.0 + (100.0 * loc[5] + 10.0 * loc[7] + 5.0) / 2400.0;
        var longitude = 2 * (10.0 * loc[0] + loc[2] - 90.0 + (100.0 * loc[4] + 10.0 * loc[6] + 5.0) / 2400.0);

        return new LatLng(latitude, longitude);
    }

 // ================================================================== UTMRef

    var UTMRef = function (easting, northing, latZone, lngZone) {
        this.easting = easting;
        this.northing = northing;
        this.latZone = latZone;
        this.lngZone = lngZone;
    };


    UTMRef.prototype.toString = function UTMRefToString() {
        return this.lngZone + this.latZone + " " + this.easting + " " + this.northing;
    };

 // ================================================================== RefEll

    var RefEll = function (maj, min) {
        this.maj = maj;
        this.min = min;
        this.ecc = ((maj * maj) - (min * min)) / (maj * maj);
    };

 // ==================================================== Conversion Functions

 /**
     * Convert a latitude and longitude into an OSGB grid reference
     *
     * @param latitudeLongitude the latitude and longitude to convert
     * @return the OSGB grid reference
     */
    LatLng.prototype.toOSRef = function LatLngToOSRef() {
        var airy1830 = new RefEll(6377563.396, 6356256.910);
        var OSGB_F0 = 0.9996012717;
        var N0 = -100000.0;
        var E0 = 400000.0;
        var phi0 = deg2rad(49.0);
        var lambda0 = deg2rad(-2.0);
        var a = airy1830.maj;
        var b = airy1830.min;
        var eSquared = airy1830.ecc;
        var phi = deg2rad(this.lat);

        var sinPhi = Math.sin(phi); // to optimize computation of trig functions and powers
        var sin2Phi = sinPhi * sinPhi;
        var cosPhi = Math.cos(phi);
        var cos2Phi = cosPhi * cosPhi;
        var cos3Phi = cos2Phi * cosPhi;
        var cos5Phi = cos3Phi * cos2Phi;
        var tanPhi = sinPhi / cosPhi;
        var tan2Phi = tanPhi * tanPhi;
        var tan4Phi = tan2Phi * tan2Phi;

        var lambda = deg2rad(this.lng);
        var E = 0.0;
        var N = 0.0;
        var n = (a - b) / (a + b);
        var xx = 1.0 - eSquared * sin2Phi;
        var v = a * OSGB_F0 * Math.pow(xx, -0.5);
        var rho = v * (1.0 - eSquared) / xx;
        var etaSquared = (v / rho) - 1.0;
        var M = (b * OSGB_F0) * (
        	((1 + n + ((5.0 / 4.0) * n * n) + ((5.0 / 4.0) * n * n * n)) * (phi - phi0)) -
            (((3 * n) + (3 * n * n) + ((21.0 / 8.0) * n * n * n)) * Math.sin(phi - phi0) * Math.cos(phi + phi0)) +
            ((((15.0 / 8.0) * n * n) + ((15.0 / 8.0) * n * n * n)) * Math.sin(2.0 * (phi - phi0)) * Math.cos(2.0 * (phi + phi0))) -
            (((35.0 / 24.0) * n * n * n) * Math.sin(3.0 * (phi - phi0)) * Math.cos(3.0 * (phi + phi0)))
        );
        var I = M + N0;
        var II = (v / 2.0) * sinPhi * cosPhi;
        var III = (v / 24.0) * sinPhi * cos3Phi * (5.0 - tan2Phi + (9.0 * etaSquared));
        var IIIA = (v / 720.0) * sinPhi * cos5Phi * (61.0 - (58.0 * tan2Phi) + tan4Phi);
        var IV = v * cosPhi;
        var V = (v / 6.0) * cos3Phi * ((v / rho) - tan2Phi);
        var VI = (v / 120.0) * cos5Phi * (
        	5.0 - (18.0 * tan2Phi) +
			(tan4Phi) + (14 * etaSquared) - (58 * tan2Phi * etaSquared)
		);

        var P = lambda - lambda0; // to optimize computation of powers
        var P2 = P * P;
        var P3 = P2 * P;
        var P4 = P2 * P2;
        var P5 = P4 * P;
        var P6 = P4 * P2;

        N = I + (II * P2) + (III * P4) + (IIIA * P6);
        E = E0 + (IV * P) + (V * P3) + (VI * P5);

        return new OSRef(E, N);
    };

 /**
     * Convert a latitude and longitude to an UTM reference
     *
     * @param ellipsoid A reference ellipsoid to use
     * @param latitudeLongitude The latitude and longitude to convert
     * @return the converted UTM reference
     */
    LatLng.prototype.toUTMRef = function LatLngToUTMRef() {
        var wgs84 = new RefEll(6378137.000, 6356752.3141);
        var UTM_F0 = 0.9996;
        var a = wgs84.maj;
        var eSquared = wgs84.ecc;
        var longitude = this.lng;
        var latitude = this.lat;

        var latitudeRad = latitude * (Math.PI / 180.0); // to optimize computation of trig functions
        var longitudeRad = longitude * (Math.PI / 180.0);
        var longitudeZone = Math.floor((longitude + 180.0) / 6.0) + 1;

 // Special zone for Norway
        if ((latitude >= 56.0) && (latitude < 64.0) && (longitude >= 3.0) && (longitude < 12.0)) {
            longitudeZone = 32;
        }

 // Special zones for Svalbard
        if ((latitude >= 72.0) && (latitude < 84.0)) {
            if ((longitude >= 0.0) && (longitude < 9.0)) {
                longitudeZone = 31;
            } else if ((longitude >= 9.0) && (longitude < 21.0)) {
                longitudeZone = 33;
            } else if ((longitude >= 21.0) && (longitude < 33.0)) {
                longitudeZone = 35;
            } else if ((longitude >= 33.0) && (longitude < 42.0)) {
                longitudeZone = 37;
            }
        }

        var longitudeOrigin = (longitudeZone - 1) * 6 - 180 + 3;
        var longitudeOriginRad = longitudeOrigin * (Math.PI / 180.0);

        var UTMZone = getUTMLatitudeZoneLetter(latitude);

        var ePrimeSquared = (eSquared) / (1 - eSquared);

        var sinLatitudeRad = Math.sin(latitudeRad);
        var cosLatitudeRad = Math.cos(latitudeRad);
        var tanLatitudeRad = sinLatitudeRad / cosLatitudeRad;

        var n = a / Math.sqrt(1 - eSquared * sinLatitudeRad * sinLatitudeRad);
        var t = tanLatitudeRad * tanLatitudeRad;
        var c = ePrimeSquared * cosLatitudeRad * cosLatitudeRad;
        var A = cosLatitudeRad * (longitudeRad - longitudeOriginRad);

        var M = a * (
        	(1 - eSquared / 4 - 3 * eSquared * eSquared / 64 - 5 * eSquared * eSquared * eSquared / 256) * latitudeRad -
            (3 * eSquared / 8 + 3 * eSquared * eSquared / 32 + 45 * eSquared * eSquared * eSquared / 1024) * Math.sin(2 * latitudeRad) +
            (15 * eSquared * eSquared / 256 + 45 * eSquared * eSquared * eSquared / 1024) * Math.sin(4 * latitudeRad) -
            (35 * eSquared * eSquared * eSquared / 3072) * Math.sin(6 * latitudeRad)
        );

        var A2 = A * A; // to optimize computation of powers
        var A3 = A2 * A;
        var A4 = A2 * A2;
        var A5 = A4 * A;
        var A6 = A5 * A;

        var UTMEasting = (UTM_F0 * n * (
        	A + (1 - t + c) * A3 / 6 +
            (5 - 18 * t + t * t + 72 * c - 58 * ePrimeSquared) * A5 / 120
        ) + 500000.0);

        var UTMNorthing = (UTM_F0 * (M + n * tanLatitudeRad * (
        	A * A / 2 + (5 - t + (9 * c) + (4 * c * c)) * A4 / 24 +
            (61 - (58 * t) + (t * t) + (600 * c) - (330 * ePrimeSquared)) * A6 / 720
        )));

 // Adjust for the southern hemisphere
        if (latitude < 0) {
            UTMNorthing += 10000000.0;
        }

        return new UTMRef(UTMEasting, UTMNorthing, UTMZone, longitudeZone);
    };

    LatLng.prototype.wgs84ToOSGB36 = function wgs84ToOSGB36() {
        var wgs84 = new RefEll(6378137.000, 6356752.3141);
        var a = wgs84.maj;
//        var b = wgs84.min;
        var eSquared = wgs84.ecc;
        var phi = deg2rad(this.lat);
        var lambda = deg2rad(this.lng);
        var v = a / (Math.sqrt(1 - eSquared * sinSquared(phi)));
        var H = 0; // height
        var cosPhi = Math.cos(phi);
        var x = (v + H) * cosPhi * Math.cos(lambda);
        var y = (v + H) * cosPhi * Math.sin(lambda);
        var z = ((1 - eSquared) * v + H) * Math.sin(phi);

        var tx = -446.448;
        var ty = 124.157;
        var tz = -542.060;
        var s = 0.0000204894;
        var rx = deg2rad(-0.00004172222);
        var ry = deg2rad(-0.00006861111);
        var rz = deg2rad(-0.00023391666);

        var xB = tx + (x * (1 + s)) + (-rx * y) + (ry * z);
        var yB = ty + (rz * x) + (y * (1 + s)) + (-rx * z);
        var zB = tz + (-ry * x) + (rx * y) + (z * (1 + s));

        var airy1830 = new RefEll(6377563.396, 6356256.910);
        a = airy1830.maj;
//        var b = airy1830.min;
        eSquared = airy1830.ecc;

        var lambdaB = rad2deg(Math.atan(yB / xB));
        var p = Math.sqrt((xB * xB) + (yB * yB));
        var phiN = Math.atan(zB / (p * (1 - eSquared)));
        var i;
        var phiN1;

        for (i = 1; i < 10; i += 1) {
            v = a / (Math.sqrt(1 - eSquared * sinSquared(phiN)));
            phiN1 = Math.atan((zB + (eSquared * v * Math.sin(phiN))) / p);
            phiN = phiN1;
        }

        var phiB = rad2deg(phiN);

        return new LatLng(phiB, lambdaB); // Note change of interface

 //this.lat = phiB;
 //this.lng = lambdaB;
    };

    LatLng.prototype.osgb36ToWGS84 = function osgb36ToWGS84() {
        var airy1830 = new RefEll(6377563.396, 6356256.910);
        var a = airy1830.maj;
//        var b = airy1830.min;
        var eSquared = airy1830.ecc;
        var phi = deg2rad(this.lat);
        var lambda = deg2rad(this.lng);
        var v = a / (Math.sqrt(1 - eSquared * sinSquared(phi)));
        var H = 0; // height
        var cosPhi = Math.cos(phi);
        var x = (v + H) * cosPhi * Math.cos(lambda);
        var y = (v + H) * cosPhi * Math.sin(lambda);
        var z = ((1 - eSquared) * v + H) * Math.sin(phi);

        var tx = 446.448;
        var ty = -124.157;
        var tz = 542.060;
        var s = -0.0000204894;
        var rx = deg2rad(0.00004172222);
        var ry = deg2rad(0.00006861111);
        var rz = deg2rad(0.00023391666);

        var xB = tx + (x * (1 + s)) + (-rx * y) + (ry * z);
        var yB = ty + (rz * x) + (y * (1 + s)) + (-rx * z);
        var zB = tz + (-ry * x) + (rx * y) + (z * (1 + s));

        var wgs84 = new RefEll(6378137.000, 6356752.3141);
        a = wgs84.maj;
//        var b = wgs84.min;
        eSquared = wgs84.ecc;

        var lambdaB = rad2deg(Math.atan(yB / xB));
        var p = Math.sqrt((xB * xB) + (yB * yB));
        var phiN = Math.atan(zB / (p * (1 - eSquared)));
        var i;
        var phiN1;

        for (i = 1; i < 10; i += 1) {
            v = a / (Math.sqrt(1 - eSquared * sinSquared(phiN)));
            phiN1 = Math.atan((zB + (eSquared * v * Math.sin(phiN))) / p);
            phiN = phiN1;
        }

        var phiB = rad2deg(phiN);

        return new LatLng(phiB, lambdaB); // Note change of interface

 //this.lat = phiB;
 //this.lng = lambdaB;
    };

    OSRef.prototype.toLatLng = function OSRefToLatLng() {
        var airy1830 = new RefEll(6377563.396, 6356256.910);
        var OSGB_F0 = 0.9996012717;
        var N0 = -100000.0;
        var E0 = 400000.0;
        var phi0 = deg2rad(49.0);
        var lambda0 = deg2rad(-2.0);
        var a = airy1830.maj;
        var b = airy1830.min;
        var eSquared = airy1830.ecc;
        var phi = 0.0;
        var lambda = 0.0;
        var E = this.easting;
        var N = this.northing;
        var n = (a - b) / (a + b);
        var M = 0.0;
        var phiPrime = ((N - N0) / (a * OSGB_F0)) + phi0;
        do {
            M = (b * OSGB_F0) * (
            	((1 + n + ((5.0 / 4.0) * n * n) + ((5.0 / 4.0) * n * n * n)) * (phiPrime - phi0)) -
            	(((3 * n) + (3 * n * n) + ((21.0 / 8.0) * n * n * n)) * Math.sin(phiPrime - phi0) * Math.cos(phiPrime + phi0)) +
            	((((15.0 / 8.0) * n * n) + ((15.0 / 8.0) * n * n * n)) * Math.sin(2.0 * (phiPrime - phi0)) * Math.cos(2.0 * (phiPrime + phi0))) -
            	(((35.0 / 24.0) * n * n * n) * Math.sin(3.0 * (phiPrime - phi0)) * Math.cos(3.0 * (phiPrime + phi0)))
            );
            phiPrime += (N - N0 - M) / (a * OSGB_F0);
        } while ((N - N0 - M) >= 0.001);

        var sinPhiPrime = Math.sin(phiPrime); // to optimize computation of trig functions and powers
        var sin2PhiPrime = sinPhiPrime * sinPhiPrime;
        var tanPhiPrime = sinPhiPrime / Math.cos(phiPrime);
        var tan2PhiPrime = tanPhiPrime * tanPhiPrime;

        var xx = 1.0 - eSquared * sin2PhiPrime;
        var v = a * OSGB_F0 * Math.pow(xx, -0.5);
        var rho = v * (1.0 - eSquared) / xx;

        var v2 = v * v; // to optimize computation of powers
        var v3 = v2 * v;
        var v5 = v3 * v2;
        var v7 = v5 * v2;

        var etaSquared = (v / rho) - 1.0;
        var VII = tanPhiPrime / (2 * rho * v);
        var VIII = (tanPhiPrime / (24.0 * rho * v3)) * (5.0 + (3.0 * tan2PhiPrime) + etaSquared - (9.0 * tan2PhiPrime * etaSquared));
        var IX = (tanPhiPrime / (720.0 * rho * v5)) * (61.0 + (90.0 * tan2PhiPrime) + (45.0 * tan2PhiPrime * tan2PhiPrime));

        var secPhiPrime = sec(phiPrime);

        var X = secPhiPrime / v;
        var XI = (secPhiPrime / (6.0 * v * v * v)) * ((v / rho) + (2 * tan2PhiPrime));
        var XII = (secPhiPrime / (120.0 * v5)) * (5.0 + (28.0 * tan2PhiPrime) + (24.0 * tan2PhiPrime * tan2PhiPrime));
        var XIIA = (secPhiPrime / (5040.0 * v7)) * (
        	61.0 + (662.0 * tan2PhiPrime) + (1320.0 * tan2PhiPrime * tan2PhiPrime) +
            (720.0 * tan2PhiPrime * tan2PhiPrime * tan2PhiPrime)
        );

        var y = E - E0; // to optimize computation of powers
        var y2 = y * y;
        var y3 = y2 * y;
        var y4 = y2 * y2;
        var y5 = y4 * y;
        var y6 = y4 * y2;
        var y7 = y6 * y;

        phi = phiPrime - (VII * y2) + (VIII * y4) - (IX * y6);
        lambda = lambda0 + (X * y) - (XI * y3) + (XII * y5) - (XIIA * y7);

        return new LatLng(rad2deg(phi), rad2deg(lambda));
    };

 /**
     * Convert an UTM reference to a latitude and longitude
     *
     * @param ellipsoid A reference ellipsoid to use
     * @param utm the UTM reference to convert
     * @return the converted latitude and longitude
     */
    UTMRef.prototype.toLatLng = function UTMRefToLatLng() {
        var wgs84 = new RefEll(6378137.000, 6356752.3141);
        var UTM_F0 = 0.9996;
        var a = wgs84.maj;
        var eSquared = wgs84.ecc;
        var ePrimeSquared = eSquared / (1.0 - eSquared);
        var e1 = (1 - Math.sqrt(1 - eSquared)) / (1 + Math.sqrt(1 - eSquared));
        var x = this.easting - 500000.0;
        var y = this.northing;
        var zoneNumber = this.lngZone;
        var zoneLetter = this.latZone;
        var longitudeOrigin = (zoneNumber - 1.0) * 6.0 - 180.0 + 3.0;

 		// Correct y for southern hemisphere
        if (zoneLetter < "N") {
            y -= 10000000.0;
        }

        var m = y / UTM_F0;
        var mu = m / (a * (
        	1.0 - eSquared / 4.0 - 3.0 * eSquared * eSquared / 64.0 -
            5.0 * Math.pow(eSquared, 3.0) / 256.0
        ));

        var phi1Rad = mu + (3.0 * e1 / 2.0 - 27.0 * Math.pow(e1, 3.0) / 32.0) * Math.sin(2.0 * mu) +
            	(21.0 * e1 * e1 / 16.0 - 55.0 * Math.pow(e1, 4.0) / 32.0) * Math.sin(4.0 * mu) +
            	(151.0 * Math.pow(e1, 3.0) / 96.0) * Math.sin(6.0 * mu);

        var sinPhi1Rad = Math.sin(phi1Rad);
        var cosPhi1Rad = Math.cos(phi1Rad);
        var tanPhi1Rad = sinPhi1Rad / cosPhi1Rad;

        var n = a / Math.sqrt(1.0 - eSquared * sinPhi1Rad * sinPhi1Rad);
        var t = tanPhi1Rad * tanPhi1Rad;
        var c = ePrimeSquared * cosPhi1Rad * cosPhi1Rad;
        var r = a * (1.0 - eSquared) / Math.pow(1.0 - eSquared * sinPhi1Rad * sinPhi1Rad, 1.5);
        var d = x / (n * UTM_F0);
        var d2 = d * d; // to optimize computation of powers
        var d3 = d2 * d;
        var d4 = d2 * d2;
        var d5 = d3 * d2;
        var d6 = d3 * d3;

        var latitude = (phi1Rad - (n * tanPhi1Rad / r) * (
        	d2 / 2.0 -
            (5.0 + (3.0 * t) + (10.0 * c) - (4.0 * c * c) - (9.0 * ePrimeSquared)) * d4 / 24.0 +
            (
            	61.0 + (90.0 * t) + (298.0 * c) + (45.0 * t * t) - (252.0 * ePrimeSquared) -
            	(3.0 * c * c)
            ) * d6 / 720.0
        )) * (180.0 / Math.PI);

        var longitude = longitudeOrigin + ((
        	d - (1.0 + 2.0 * t + c) * d3 / 6.0 +
            (
            	5.0 - (2.0 * c) + (28.0 * t) - (3.0 * c * c) + (8.0 * ePrimeSquared) +
            	(24.0 * t * t)
            ) * d5 / 120.0
        ) / cosPhi1Rad) * (180.0 / Math.PI);

        return new LatLng(latitude, longitude);
    };

    var itself = {};

    itself.LatLng = LatLng;
    itself.OSRef = OSRef;
    itself.UTMRef = UTMRef;

    itself.getOSRefFromOSString = getOSRefFromOSString;
    itself.getOSRefFromSixFigureReference = getOSRefFromOSString;

    itself.getLatLngFromLocator = getLatLngFromLocator;

    return itself;

}());

// ================================================== End of Module
