/*
	NGR Converter - National Grid Reference to Latitude and Longitude
 	Copyright © 2005-2019 Harry Whitfield

 	This program is free software; you can redistribute it and/or modify it
 	under the terms of the GNU General Public License as published by the
 	Free Software Foundation; either version 2 of the License, or (at your
 	option) any later version.

 	This program is distributed in the hope that it will be useful, but
 	WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 	General Public License for more details.

 	You should have received a copy of the GNU General Public License along
 	with this program; if not, write to the Free Software Foundation, Inc.,
 	51 Franklin St, Fifth Floor, Boston, MA	 02110-1301	 USA

 	NGR Converter - browser version 2.1
 	29 April, 2019
 	Copyright © 2005-2019 Harry Whitfield
 	mailto:g6auc@arrl.net
*/

/*jslint browser, multivar, this */

/*property
    PI, abs, cos, eastings, latitude, longitude, northings, sin, sqrt, tan
*/

// Airy 1830: a = 6377563.396, b = 6356256.910
var n = 0.001673220250;
var e2 = 0.006670539762;
var a = 6375020.481;	// 6377563.396 x f0 f0 = 0.9996012717;
var b = 6353722.490;	// 6356256.910 x f0
var f0 = 0.9996012717;

var eorg = 400000.0;
var norg = -100000.0;

var pi = Math.PI;

var n2 = n * n;
var n3 = n * n * n;

var korg = 49.0 * pi / 180.0;	// in radians
var lorg = -2.0 * pi / 180.0;	// in radians

function GridRef(eastings, northings) {
	"use strict";
	this.eastings = eastings;
	this.northings = northings;
}

function Position(latitude, longitude) {
	"use strict";
	this.latitude = latitude;
	this.longitude = longitude;
}

function fnm(k3, k4) {	// k3, k4 in radians
	"use strict";
	var j3 = (1 + n + 1.25 * n2 + 1.25 * n3) * k3,
		j4 = (3 * n + 3 * n2 + 2.625 * n3) * Math.sin(k3) * Math.cos(k4),
		j5 = (1.875 * n2 + 1.875 * n3) * Math.sin(2 * k3) * Math.cos(2 * k4),
		j6 = (35.0 / 24.0) * n3 * Math.sin(3 * k3) * Math.cos(3 * k4);

	return (b * (j3 - j4 + j5 - j6));
}

function fnv(k) {	// k in radians
	"use strict";
	var s = Math.sin(k);

	return (a / Math.sqrt(1 - e2 * s * s));
}

function fnr(v, k) {
	"use strict";
	var s = Math.sin(k);

	return (v * (1 - e2) / (1 - e2 * s * s));
}

function ngr(k, l) { // lat,long -> eastings,northings	k, l in radians
	"use strict";
	var m = fnm(k - korg, k + korg),
		v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		s = Math.sin(k),
		c = Math.cos(k),
		t = Math.tan(k),
		t2 = t * t,
		t4 = t2 * t2,
		c2 = c * c,
		c3 = c * c2,
		c5 = c2 * c3,
		p = l - lorg,
		j3 = m + norg,
		j4 = 0.5 * v * s * c,
		j5 = (v / 24) * s * c3 * (5 - t2 + 9 * h2),
		j6 = (v / 720) * s * c5 * (61 - 58 * t2 + t4),
		northings = j3 + p * p * j4 + p * p * p * p * j5 + p * p * p * p * p * p * j6,
		j7 = v * c,
		j8 = (v / 6) * c3 * (v / r - t2),
		j9 = (v / 120) * c5 * (5 - 18 * t2 + t4 + 14 * h2 - 58 * t2 * h2),
		eastings = eorg + p * j7 + p * p * p * j8 + p * p * p * p * p * j9;
	return new GridRef(eastings, northings);
}

function phi(n) {
	"use strict";
	var k = korg,
		m = 0;

	do {
		k = k + (n - norg - m) / a;
		m = fnm(k - korg, k + korg);
	} while (Math.abs(n - norg - m) >= 0.001);
	return k;
}

function lat_long(e, n) { // ngr -> latitude,longitude
	"use strict";
	var k = phi(n),
		c = Math.cos(k),
		t = Math.tan(k),
		t2 = t * t,
		t4 = t2 * t2,
		t6 = t2 * t4,
		v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		y = e - eorg,
		sV = 0.000001 * v,
		sR = 0.000001 * r,
		sy = 0.000001 * y,
		j3 = t / (2 * sR * sV),
		j4 = t / (24 * sR * sV * sV * sV) * (5 + 3 * t2 + h2 - 9 * t2 * h2),
		j5 = t / (720 * sR * sV * sV * sV * sV * sV) * (61 + 90 * t2 + 45 * t4),
		k9 = k - sy * sy * j3 + sy * sy * sy * sy * j4 - sy * sy * sy * sy * sy * sy * j5,
		j6 = 1.0 / (c * sV),
		j7 = 1.0 / (c * 6 * sV * sV * sV) * (sV / sR + 2 * t2),
		j8 = 1.0 / (c * 120 * sV * sV * sV * sV * sV) * (5 + 28 * t2 + 24 * t4),
		j9 = 1.0 / (c * 5040 * sV * sV * sV * sV * sV * sV * sV),
		sj9 = j9 * (61 + 662 * t2 + 1320 * t4 + 720 * t6),
		ll = lorg + sy * j6 - sy * sy * sy * j7 + sy * sy * sy * sy * sy * j8 - sy * sy * sy * sy * sy * sy * sy * sj9,
		kk = k9;

	return new Position(kk, ll);	// kk, ll in radians
}

function fnc(k, l) {	// k, l in radians
	"use strict";
	var v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		s = Math.sin(k),
		c = Math.cos(k),
		t = Math.tan(k),
		t2 = t * t,
		c2 = c * c,
		c4 = c2 * c2,
		p = l - lorg,
		j4 = s * c2 / 3.0 * (1 + 3 * h2 + 2 * h2 * h2),
		j5 = s * c4 / 15.0 * (2 - t2);

	return (p * s + p * p * p * j4 + p * p * p * p * p * j5);
}

function con(e, n) {
	"use strict";
	var k = phi(n),
		t = Math.tan(k),
		t2 = t * t,
		t4 = t2 * t2,
		v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		y = e - eorg,
		sV = 0.000001 * v,
//		sR = 0.000001 * r,
		sy = 0.000001 * y,
		j3 = t / sV,
		j4 = t / (3 * sV * sV * sV) * (1 + t2 - h2 - 2 * h2 * h2),
		j5 = t / (15 * sV * sV * sV * sV * sV) * (2 + 5 * t2 + 3 * t4);

	return (sy * j3 - sy * sy * sy * j4 + sy * sy * sy * sy * sy * j5);
}

function fnf(k, l) {	// k, l in radians
	"use strict";
	var v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		c = Math.cos(k),
		t = Math.tan(k),
		t2 = t * t,
		c2 = c * c,
		c4 = c2 * c2,
		p = l - lorg,
		j3 = c2 / 2.0 * (1 + h2),
		j4 = c4 / 24.0 * (5 - 4 * t2 + 14 * h2 - 28 * t2 * h2);

	return (f0 * (1 + p * p * j3 + p * p * p * p * j4));
}

function fac(e, n) {
	"use strict";
	var k = phi(n),
		v = fnv(k),
		r = fnr(v, k),
		h2 = v / r - 1.0,
		y = e - eorg,
		sV = 0.000001 * v,
		sR = 0.000001 * r,
		sy = 0.000001 * y,
		j3 = 1.0 / (2 * sR * sV),
		j4 = (1 + 4 * h2) / (24 * sR * sR * sV * sV);

	return (f0 * (1 + sy * sy * j3 + sy * sy * sy * sy * j4));
}

function deltaTa(ea, na, eb, nb) {
	"use strict";
	var k = phi(0.5 * (na + nb)),
		v = fnv(k),
		r = fnr(v, k),
//		h2 = v / r - 1.0,
		ya = ea - eorg,
		yb = eb - eorg,
		sV = 0.001 * v,
		sR = 0.001 * r,
		sya = 0.000001 * ya,
		syb = 0.000001 * yb,
		j3 = 1.0 / (6 * sR * sV);

	return ((2 * sya + syb) * (na - nb) * j3);
}

function deltaTb(ea, na, eb, nb) {
	"use strict";
	var k = phi(0.5 * (na + nb)),
		v = fnv(k),
		r = fnr(v, k),
//		h2 = v / r - 1.0,
		ya = ea - eorg,
		yb = eb - eorg,
		sV = 0.001 * v,
		sR = 0.001 * r,
		sya = 0.000001 * ya,
		syb = 0.000001 * yb,
		j3 = 1.0 / (6 * sR * sV);

	return ((sya + 2 * syb) * (nb - na) * j3);
}

/*
// Test code

var pos = ngr(53.65788888 * pi / 180, -1.9466944444 * pi / 180);
print(pos.eastings.toFixed(3), pos.northings.toFixed(3));

var loc = lat_long(403522.401, 417960.014);
var lat = loc.latitude * 180 / pi;
var lng = loc.longitude * 180 / pi;
print(lat.toFixed(10), lng.toFixed(10));
*/