/*
	NGR Converter - National Grid Reference to Latitude and Longitude
 	Copyright © 2005-2019 Harry Whitfield

 	This program is free software; you can redistribute it and/or modify it
 	under the terms of the GNU General Public License as published by the
 	Free Software Foundation; either version 2 of the License, or (at your
 	option) any later version.

 	This program is distributed in the hope that it will be useful, but
 	WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 	General Public License for more details.

 	You should have received a copy of the GNU General Public License along
 	with this program; if not, write to the Free Software Foundation, Inc.,
 	51 Franklin St, Fifth Floor, Boston, MA	 02110-1301	 USA

 	NGR Converter - browser version 2.1
 	29 April, 2019
 	Copyright © 2005-2019 Harry Whitfield
 	mailto:g6auc@arrl.net
*/

/*jslint browser, multivar */

/*global paramOne, paramTwo, paramThree, paramFour,
	headingOne, headingTwo, headingThree, headingFour, clearData, Menu,

	GridRef, Position, get_ref, put_ref, lat_long, put_lat, put_long,
	put_locator, put_convergence, put_scale, get_ang, ngr, get_locator,
	JScoord, setPreference, getPreferences, cookieAlert
*/

/*property
    LatLng, PI, eastings, isNaN, keyCode, lat, latitude, lng, longitude,
    northings, onchange, onkeypress, osgb36ToWGS84, selectedIndex, title,
    value, wgs84ToOSGB36, which
*/

var gWidgetName;	// global

// map custom names onto model names
var reference = paramOne;
var latitude = paramTwo;
var longitude = paramThree;
var locator = paramFour;

var referenceHeading = headingOne;
var latitudeHeading = headingTwo;
var longitudeHeading = headingThree;
var locatorHeading = headingFour;

var refModeMenu = Menu[0];
var digitsMenu = Menu[1];
var latLongMenu = Menu[2];
var locatorMenu = Menu[3];
var datumMenu = Menu[4];

var prec1 = 4; // precision for seconds of arc
var prec2 = 3; // precision for eastings, northings
var prec3 = 8; // precision for convergence
var prec4 = 7; // precision for numeric latitude/longitude

var numericRefsPref;	// = "1"
var figuresPref;		// = "10"
var latLongPref;		// = "0";
var locatorPref;		// = "IARU Locator";
var datumPref;			// = "OSGB36";

/*
    ["Alphanumeric", "Numeric"],
    ["2", "4", "6", "8", "10"],
    ["Formatted", "Numeric"],
    ["IARU Locator", "Convergence", "Scale"],
    ["OSGB36", "WGS84"]
*/

var E = 400000;
var N = -100000;
var G = new GridRef(E, N);

var K = 49.0;
var L = -2.0;
var P = new Position(K, L);

var item = [];

function deg2rad(x) {
    "use strict";
    return x * (Math.PI / 180);
}

function rad2deg(x) {
    "use strict";
    return x * (180 / Math.PI);
}

function convertToWGS84(pp) {	// pp is a Position
    "use strict";
	var qq = new JScoord.LatLng(rad2deg(pp.latitude), rad2deg(pp.longitude));
	var rr = qq.osgb36ToWGS84();
	return new Position(deg2rad(rr.lat), deg2rad(rr.lng));
}

function convertToOSGB36(pp) {	// pp is a Position
    "use strict";
	var qq = new JScoord.LatLng(rad2deg(pp.latitude), rad2deg(pp.longitude));
	var rr = qq.wgs84ToOSGB36();
	return new Position(deg2rad(rr.lat), deg2rad(rr.lng));
}

function setMenus() {
    "use strict";
	refModeMenu.value = (
		numericRefsPref === "1"
		? "Numeric"
		: "Alphanumeric"
	);
	digitsMenu.value = figuresPref;
	latLongMenu.value = (
		latLongPref === "1"
		? "Numeric"
		: "Formatted"
	);
	locatorMenu.value = locatorPref;
	datumMenu.value = datumPref;
}

function updateValues() {
    "use strict";
    var g, p;

    if (reference.value !== "") {
        g = get_ref(reference.value);
        if (Number.isNaN(g.eastings)) {
            reference.value = "Invalid Input";
            latitude.value = "";
            longitude.value = "";
            locator.value = "";
        } else {
        	reference.value = put_ref(g);
            p = lat_long(g.eastings, g.northings);
            if (datumPref === "WGS84") {
            	p = convertToWGS84(p);
            }
            latitude.value = put_lat(p.latitude);
            longitude.value = put_long(p.longitude);
            if (locatorPref === "IARU Locator") {
            	if (datumPref === "OSGB36") {
            		p = convertToWGS84(p);
            	}
                locator.value = put_locator(p.latitude, p.longitude);
            }
            if (locatorPref === "Convergence") {
            	if (datumPref === "WGS84") {
            		p = convertToOSGB36(p);
            	}
                locator.value = put_convergence(p.latitude, p.longitude);
            }
            if (locatorPref === "Scale") {
            	if (datumPref === "WGS84") {
            		p = convertToOSGB36(p);
            	}
				locator.value = put_scale(p.latitude, p.longitude);
            }
        }
    }

}

function updatePrefs() {
    "use strict";
    if (locatorPref === "IARU Locator") {
        locator.title = "Enter a 6 or 8 character IARU locator in this box.";
        locatorHeading.value = locatorMenu.value + " (WGS84)";
    }
    if (locatorPref === "Convergence") {
        locator.title = "Computed Convergence.";
        locatorHeading.value = locatorMenu.value + " (OSGB36)";
    }
    if (locatorPref === "Scale") {
        locator.title = "Computed Scale.";
        locatorHeading.value = locatorMenu.value + " (OSGB36)";
    }

    if (datumPref === "OSGB36") {
    	latitudeHeading.value = "Latitude (OSGB36)";
    	longitudeHeading.value = "Longitude (OSGB36)";
    } else {
    	latitudeHeading.value = "Latitude (WGS84)";
    	longitudeHeading.value = "Longitude (WGS84)";
    }

	updateValues();
}

function test_case(g) {
    "use strict";
    var p = lat_long(g.eastings, g.northings);
    if (datumPref === "WGS84") {
    	p = convertToWGS84(p);
    }
    reference.value = put_ref(g);
    latitude.value = put_lat(p.latitude);
    longitude.value = put_long(p.longitude);
    if (locatorPref === "IARU Locator") {
        if (datumPref === "OSGB36") {
            p = convertToWGS84(p);
        }
        locator.value = put_locator(p.latitude, p.longitude);
    }
    if (locatorPref === "Convergence") {
        if (datumPref === "WGS84") {
            p = convertToOSGB36(p);
        }
        locator.value = put_convergence(p.latitude, p.longitude);
    }
    if (locatorPref === "Scale") {
        if (datumPref === "WGS84") {
            p = convertToOSGB36(p);
        }
		locator.value = put_scale(p.latitude, p.longitude);
    }
}

refModeMenu.onchange = function () {
    "use strict";
	numericRefsPref = (
		refModeMenu.selectedIndex === 1
		? "1"
		: "0"
	);
    setPreference("numericRefsPref", numericRefsPref);
	updatePrefs();
};

digitsMenu.onchange = function () {
    "use strict";
	figuresPref = digitsMenu.value;
	setPreference("figuresPref", figuresPref);
	updatePrefs();
};

latLongMenu.onchange = function () {
    "use strict";
	latLongPref = (
		latLongMenu.selectedIndex === 1
    	? "1"
    	: "0"
    );
	setPreference("latLongPref", latLongPref);
	updatePrefs();
};

locatorMenu.onchange = function () {
    "use strict";
	locatorPref = locatorMenu.value;
	locatorHeading.value = locatorMenu.value;
	setPreference("locatorPref", locatorPref);
	updatePrefs();
};

datumMenu.onchange = function () {
    "use strict";
	datumPref = datumMenu.value;
	setPreference("datumPref", datumPref);
	updatePrefs();
};

reference.onkeypress = function (event) {
    "use strict";
	var x = event.which || event.keyCode;

	if (x === 13) {
		updateValues();
    } else {
        latitude.value = "";
        longitude.value = "";
        locator.value = "";
    }
};

latitude.onkeypress = function (event) {
    "use strict";
	var x = event.which || event.keyCode;
	var pp;

	if (x === 13) {
        if (latitude.value !== "") {
            K = get_ang(latitude.value, 90, "N", "S");
            L = get_ang(longitude.value, 180, "E", "W");

            if (Number.isNaN(K) || Number.isNaN(L)) {
                reference.value = "Invalid Input";
                locator.value = "";
            } else {
				if (datumPref === "WGS84") {
            		pp = new Position(K, L);
            		pp = convertToOSGB36(pp);
            		K = pp.latitude;
            		L = pp.longitude;
            	}
                G = ngr(K, L);
                reference.value = put_ref(G);
                if (locatorPref === "IARU Locator") {
					if (datumPref === "OSGB36") {
            			pp = new Position(K, L);
            			pp = convertToWGS84(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_locator(K, L);
                }
                if (locatorPref === "Convergence") {
            		if (datumPref === "WGS84") {
            			pp = new Position(K, L);
            			pp = convertToOSGB36(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_convergence(K, L);
                }
                if (locatorPref === "Scale") {
            		if (datumPref === "WGS84") {
            			pp = new Position(K, L);
            			pp = convertToOSGB36(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_scale(K, L);
                }
            }
        }
    } else {
        reference.value = "";
        locator.value = "";
    }
};

longitude.onkeypress = function (event) {
    "use strict";
	var x = event.which || event.keyCode;
	var pp;

	if (x === 13) {
        if (longitude.value !== "") {
            K = get_ang(latitude.value, 90, "N", "S");
            L = get_ang(longitude.value, 180, "E", "W");

            if (Number.isNaN(K) || Number.isNaN(L)) {
                reference.value = "Invalid Input";
                locator.value = "";
            } else {
				if (datumPref === "WGS84") {
            		pp = new Position(K, L);
            		pp = convertToOSGB36(pp);
            		K = pp.latitude;
            		L = pp.longitude;
            	}
                G = ngr(K, L);
                reference.value = put_ref(G);
                if (locatorPref === "IARU Locator") {
					if (datumPref === "OSGB36") {
            			pp = new Position(K, L);
            			pp = convertToWGS84(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_locator(K, L);
                }
                if (locatorPref === "Convergence") {
            		if (datumPref === "WGS84") {
            			pp = new Position(K, L);
            			pp = convertToOSGB36(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_convergence(K, L);
                }
                if (locatorPref === "Scale") {
            		if (datumPref === "WGS84") {
            			pp = new Position(K, L);
            			pp = convertToOSGB36(pp);
            			K = pp.latitude;
            			L = pp.longitude;
            		}
                    locator.value = put_scale(K, L);
                }
            }
        }
    } else {
        reference.value = "";
        locator.value = "";
    }
};

locator.onkeypress = function (event) {
    "use strict";
	var x = event.which || event.keyCode;
	var pp;

	if (x === 13) {
        if (locator.value !== "") {
            P = get_locator(locator.value);
            if (Number.isNaN(P.latitude)) {
                reference.value = "Invalid Input";
                latitude.value = "";
                longitude.value = "";
            } else {
            	if (datumPref === "OSGB36") {
            		pp = new Position(P.latitude, P.longitude);
            		pp = convertToOSGB36(pp);
            		P.latitude = pp.latitude;
            		P.longitude = pp.longitude;
            	}
                latitude.value = put_lat(P.latitude);
                longitude.value = put_long(P.longitude);

            	if (datumPref === "WGS84") {
            		pp = new Position(P.latitude, P.longitude);
            		pp = convertToOSGB36(pp);
            		P.latitude = pp.latitude;
            		P.longitude = pp.longitude;
            	}
                G = ngr(P.latitude, P.longitude);
                reference.value = put_ref(G);
            }
        }
    } else {
        reference.value = "";
        latitude.value = "";
        longitude.value = "";
    }
};

gWidgetName = "NGR Converter Web Widget\n\n";
getPreferences();
setMenus();
updatePrefs();
test_case(new GridRef(651409.903, 313177.270));
